#include <iostream>

// classes, inheritance and abstract classes

/* first simple class */
class A {
  public:
    A() : _i(), _a() {} // default constructor if no constructor specified
    A (int i, char a) : _i(i), _a(a) {} // direct initialization
  private:
    int  _i;
    char _a;
}; // note ";" after class definition


/* more types of constructors */
class B {
  public:
    B (int i, char a) : _i(i), _a(a) {
      std::cout << "B: direct initialization" << std::endl;
    }    // direct initialization

    B (const B& b) : _i(b._i), _a(b._a) {
      std::cout << "B: copy constructor" << std::endl;
    } // copy constructor

    B& operator= (const B& b) {
      std::cout << "B: assignment operator" << std::endl;
      _i = b._i;
      _a = b._a;
      return *this;
    }

    // Destructor
    ~B () {
      std::cout << "B: destructor" << std::endl;
    }

  private:
    int  _i;
    char _a;
};


/* 
 * Inheritance
 * Virtual functions allow for dynamic binding, like it is standard in Java.
 * See main method below for more information.
 */
class Base {
  public:
    void func () {
      std::cout << "base: func" << std::endl;
    }
    void nonVirtFunc () {
      std::cout << "base: non-virtual function" << std::endl;
    }
    virtual void virtFunc () {
      std::cout << "base: virtual function" << std::endl;
    }
};


class Derived : public Base {
  public:
    void nonVirtFunc () {
      std::cout << "derived: non-virtual function" << std::endl;
    }
    void virtFunc () override { // override keyword is not a must, but recommended
      std::cout << "derived: overriden virtual function" << std::endl;
    }
};


/* 
 * Abstract Class
 * An abstract class is one with at least one pure virtual function
 * pure virtual functions have no implementation, instead they have = 0.
 * You cannot instantiate an abstract class. 
 */
class AbstractBase {
  public:
    virtual void doSomething () = 0;
};

class Derived2 : public AbstractBase {
  public:
    void doSomething () override {
      std::cout << "doSomething" << std::endl;
    }
};


int main () {
  std::cout << "=== Program start ===" << std::endl;

  // instantiate object with default constructor.
  // note: no parentheses () allowed here
  A a;
  //A a (); // error: for weird reasons the compiler interpretes this as a function decleration

  // Calls to different constructors
  std::cout << "\n*** Constructor calls to B ***" << std::endl;
  B b1 (1, 'a');  // direct initialization
  B b2 (b1);      // copy initialization
  b1 = b2;        // assignment


  // Inheritance and Polymorphism
  std::cout << "\n*** Inheritance & Polymorphism" << std::endl;
  Derived derived;
  derived.func();
  Base* basePointer = &derived; // polymorphism
  basePointer->virtFunc(); // virtual: dynamic binding, make look-up at runtime
  basePointer->nonVirtFunc(); // non-virtual: static binding


  // Usage of an abstract class
  //AbstractBase abstractBase; // error: cannot declare variable to be of abstract type
  Derived2 derived2;
  AbstractBase* abstractBasePointer = &derived2; // pointer is ok
  (void) abstractBasePointer; // Silence compiler warning

  std::cout << "\n=== Program end ===" << std::endl;
}