#include <iostream> 
// for printing variables: std::cout << {{variableName}} << std::endl;

// Function to print an array
void print (char* A, unsigned Asize) {
    for (unsigned i = 0; i < Asize; ++i) {
        std::cout << A[i] << "  ";
    }
    std::cout << std::endl;
}


// Function: call by value
double funcByVal (double d) {
  d *= 2;
  return d;
}


// Function: call by reference
double funcByRef (double& d) {
  d *= 2;
  return d;
}


// Function: call by reference using pointer
double funcByPointer (double* d) {
  *d *= 2;
  return *d;
}


// main method: program starts here
int main () {

  std::cout << "=== Program start ===" << std::endl;

  /* *** 1. Pointers and References *** */
  std::cout << "\n*** 1. Pointers and References ***" << std::endl;
  int i = 5;
  int* ip; // int pointer, pointers store address of variable
  ip = &i; // ip holds address of i

  // use *ip to dereference (get value of variable at address) pointer
  std::cout << "value of integer at address " << ip << " is: " << *ip << std::endl;
  // int& ir; // ERROR: a reference must be initialized
  int& ir = i; // ir references i. ir serves as another name for i
  
  int i2 = 3;
  ip = &i2; // now ip holds address of i2
  ir = i2; // does not mean that ir now references i2, but that i now has the value of i2
  std::cout << "i: " << i << "  |  i2: " << i2 << "  |  ir: " << ir << std::endl;
  
  int** ipp = & ip; // ipp holds address of ip
  std::cout << "ip: " << ip << "  |  &ip: " << &ip << "  |  ipp: " << ipp << std::endl;
  

  /* *** 2. Const *** */
  std::cout << "\n*** 2. Const ***" << std::endl;
  double        d   = 3.4;
  double        d2  = 1.8;
  const double  cd  = 0.5;
  // cd += 2;              // error: trying to assign const variable
  
  const double& cdr = d;   // value of d cannot be changed through reference cdr
  d     = 7.0;
  //cdr = 7.0;             // error: assignment of read-only reference

  const double* cdp = &d;  // pointer of d cannot be changed through pointer dp
  //*cdp = 5.0;            // error: assignment of read-only location
  cdp = &d2;               // but dp can point to other variable

  double* const dcp = &d;  // pointer cannot point to another variable
  *dcp = 5.0;
  //dcp  = &d2;            // error: assignment of read-only variable

  const double* const cdcp = &d; // guess what

  
  /* *** 3. Arrays *** */
  std::cout << "\n*** 3. Arrays ***" << std::endl;
  
  char A [] = {'a', 'b', 'c', 'd'}; // char array
  std::cout << "size of A in bytes:      " << sizeof(A) << std::endl;
  std::cout << "number of elements in A: " << sizeof(A) / sizeof(char) << std::endl;
  
  char* cp = &A[0]; // pointer to first element. alternative: cp = A
  cp += 2;
  std::cout << *cp << std::endl;
  
  // when passing an array to a function: pass pointer to first element and size
  print(A, sizeof(A)/sizeof(char));
  // equivalent: print(&A[0], sizeof(A)/sizeof(char));


  /* *** 4. Call by Value & Call by Reference *** */
  std::cout << "\n*** 4. Call by Value & Call by Reference ***" << std::endl;
  
  double x = 5.0;
  double funcRes = 0.0;

  funcRes = funcByVal(x);
  std::cout << "funcByVal returns:     " << funcRes << "  |  x: " << x << std::endl;

  funcRes = funcByRef(x);
  std::cout << "funcByRef returns:     " << funcRes << "  |  x: " << x << std::endl;

  funcRes = funcByPointer(&x);
  std::cout << "funcByPointer returns: " << funcRes << "  |  x: " << x << std::endl;


  std::cout << "\n=== Program end ===" << std::endl;

  return 0;
}
